"""
Functions related to the CORSIKA longitudinal distribution
"""
import re
import numpy as np

PARTICLE_HEADER_RE = re.compile(r"LONGITUDINAL DISTRIBUTION IN\s+(\d+)\s+(SLANT|VERTICAL)\s+STEPS OF\s+(\d+(?:.\d*)?) G\/CM\*\*2 FOR SHOWER\s+(\d+)")
ENERGY_HEADER_RE = re.compile(r"LONGITUDINAL ENERGY DEPOSIT IN\s+(\d+)\s+(SLANT|VERTICAL)\s+STEPS OF\s+(\d+(?:.\d*)?) G\/CM\*\*2 FOR SHOWER\s+(\d+)")

ENERGY_COLUMNS = [
    "depth",
    "gamma", "em_ioniz", "em_cut", "mu_ioniz",
    "mu_cut", "hadr_ioniz", "hadr_cut", "neutrino", "sum"
]
PARTICLE_COLUMNS = [
    "depth",
    "gammas", "positrons", "electrons", "mu_plus", "mu_minus",
    "hadrons", "charged", "nuclei", "cherenkov"
]


def longitudinal_fit_function(depth, n_max, depth_0, depth_max, a, b, c):
    """
    The function CORSIKA fits to the longitudinal distribution.

    Parameters of this function are stored in the event end block
    in case the corresponding options of the LONGI keyword are set.

    Use like this:

    >>> depth = np.linspace(0, 1000, 100)
    >>> longitudinal_fit_function(depth, *event.end["longitudinal_fit_parameters"])
    """
    denominator =  (a + b * depth + c * depth**2)
    exponent = (depth_max - depth_0) / denominator
    power = ((depth - depth_0) / (depth_max - depth_0))**exponent
    exp = np.exp((depth_max - depth) / denominator)
    return n_max * power *  exp


def read_longitudinal_distributions(path):
    """
    Read longitudinal profiles from CORSIKA longitudinal file.

    This function returns a generator that iterates over air showers.

    Parameters
    ----------
    path : str, Path
        Path to CORSIKA longitudinal output file (DATXXXXXX.long)

    Yields
    ------
    longitudinal : dict
        Dict with the information for one air shower.
        Contains the longitudinal tables for "particles" and "energy_deposition"
        and the "parameters", "chi2_ndf" and "average_deviation" values of the 
        fit to the distribution if available.
    """
    first = True
    with open(path, "r") as f:
        try:
            line = f.readline().strip()
        except UnicodeDecodeError:
            raise IOError(f"Inputfile {path} does not seem to be a longitudinal file")

        while line:
            match = PARTICLE_HEADER_RE.match(line)
            if not match:
                if first:
                    raise IOError(f"Inputfile {path} does not seem to be a longitudinal file")
                else:
                    raise IOError(f"Error reading file, expected header line, got: {line}")
            first = False

            n_steps = int(match.group(1))
            longi = dict(
                shower=int(match.group(4)),
                n_steps=n_steps,
                slant=match.group(2) == "SLANT",
                step_width=float(match.group(3)),
            )

            # skip line with names
            f.readline()
            longi["particles"] = np.genfromtxt(
                f, max_rows=n_steps, names=PARTICLE_COLUMNS, dtype=None
            )

            line = f.readline().strip()
            match = ENERGY_HEADER_RE.match(line)
            if not match:
                raise IOError(f"Error reading file, expected energy deposition header line, got: {line}")
            n_steps = int(match.group(1))

            # skip
            f.readline()
            longi["energy_deposition"] = np.genfromtxt(f, max_rows=n_steps, names=ENERGY_COLUMNS, dtype=None)

            line = f.readline()
            if line.strip().startswith("FIT"):
                f.readline()
                parameters = f.readline().partition(" = ")[2]
                longi["parameters"] = np.array([float(p) for p in parameters.split()])

                chi2_ndf = f.readline().partition(" = ")[2]
                longi["chi2_ndf"] = float(chi2_ndf)

                deviation = f.readline().partition(" = ")[2]
                longi["average_deviation"] = float(deviation)

            f.readline()
            yield longi
            line = f.readline().strip()



from functools import cache
import numpy as np

from .dtypes import build_dtype, Field, filter_fields_by_version

event_end_fields = [
    Field(1, "event_end", dtype="S4", min_version=np.float32(6.5)),
    Field(2, "event_number", min_version=np.float32(6.5)),
    Field(3, "n_photons_weighted", min_version=np.float32(6.5)),
    Field(4, "n_electrons_weighted", min_version=np.float32(6.5)),
    Field(5, "n_hadrons_weighted", min_version=np.float32(6.5)),
    Field(6, "n_muons_weighted", min_version=np.float32(6.5)),
    Field(7, "n_particles_written", min_version=np.float32(6.5)),
    Field(256, "longitudinal_fit_parameters", shape=6, min_version=np.float32(7.3)),
    Field(262, "chi_square_longitudinal", min_version=np.float32(6.5)),
    Field(263, "n_photons_written", min_version=np.float32(6.5)),
    Field(264, "n_electrons_written", min_version=np.float32(6.5)),
    Field(265, "n_hadrons_written", min_version=np.float32(6.5)),
    Field(266, "n_muons_written", min_version=np.float32(6.5)),
    Field(267, "n_em_particles_preshower", min_version=np.float32(6.5)),
]


def get_event_end_fields(version):
    return filter_fields_by_version(event_end_fields, version)


@cache
def get_event_end_dtype(version):
    return build_dtype(get_event_end_fields(version))


@cache
def get_event_end_thin_dtype(version):
    return build_dtype(get_event_end_fields(version), itemsize=4 * 312)



from collections import namedtuple, defaultdict
import numpy as np


Field = namedtuple(
    "Field",
    ["position", "name", "unit", "shape", "dtype", "min_version"],
    defaults=(None, 1, "f4", np.float32(0)),
)


def build_dtype(fields, itemsize=4 * 273):
    dt = defaultdict(list)
    if itemsize is not None:
        dt["itemsize"] = itemsize

    for field in fields:
        dt["names"].append(field.name)
        dt["offsets"].append((field.position - 1) * 4)
        if field.shape != 1:
            dt["formats"].append((field.dtype, field.shape))
        else:
            dt["formats"].append(field.dtype)

    return np.dtype(dict(**dt))


def filter_fields_by_version(fields, version):
    """Return only fields whose min_version is <= the provided version."""
    return [field for field in fields if field.min_version <= version]



import gzip
import struct
from enum import Enum

from .constants import BLOCK_SIZE_BYTES, BLOCK_SIZE_BYTES_THIN


#: buffersize for files not written as fortran sequential file
DEFAULT_BUFFER_SIZE = BLOCK_SIZE_BYTES * 100
DEFAULT_BUFFER_SIZE_THIN = BLOCK_SIZE_BYTES_THIN * 100
#: struct definition of the fortran record marker
RECORD_MARKER = struct.Struct('i')


class MagicBytes(Enum):
    GZIP = b"\x1f\x8b"
    ZSTD = b"\x28\xb5\x2f\xfd"
    EVENTIO_LE = b"\x37\x8a\x1f\xd4"
    EVENTIO_BE = b"\xd4\x1f\x8a\x37"


MAX_MARKER_LENGTH = max(len(marker.value) for marker in MagicBytes)


def check_magic(magic, *, path=None, marker_bytes=None):
    if (path is None) == (marker_bytes is None):
        raise ValueError("Need to provide exactly one of path xor marker_bytes")

    if marker_bytes is None:
        with open(path, "rb") as f:
            marker_bytes = f.read(MAX_MARKER_LENGTH)

    expected = magic.value
    return marker_bytes[: len(expected)] == expected


def is_gzip(path):
    """Test if a file is gzip compressed."""
    return check_magic(MagicBytes.GZIP, path=path)


def is_zstd(path):
    """Test if a file is zstd compressed."""
    return check_magic(MagicBytes.ZSTD, path=path)


class ClosingGzipFile(gzip.GzipFile):
    """A wrapper around std lib GzipFile that auto-closes the underlying fobj."""

    def close(self):
        if self.fileobj is not None:
            self.fileobj.close()
        super().close()


def open_compressed(path):
    fobj = open(path, "rb")
    marker_bytes = fobj.read(MAX_MARKER_LENGTH)
    fobj.seek(0)

    if check_magic(MagicBytes.GZIP, marker_bytes=marker_bytes):
        return ClosingGzipFile(mode="rb", fileobj=fobj)

    if check_magic(MagicBytes.ZSTD, marker_bytes=marker_bytes):
        from zstandard import ZstdDecompressor

        return ZstdDecompressor().stream_reader(fobj, closefd=True)

    return fobj


def read_buffer_size(path):
    '''
    Reads the first 4 bytes of a file and checks if
    it is the 'RUNH' designation None is returned,
    if not interpret it as unsigned integer, the
    size of the CORSIKA buffer in bytes
    '''
    with open_compressed(path) as f:
        data = f.read(RECORD_MARKER.size)

        if data == b'RUNH':
            return None

        buffer_size, = RECORD_MARKER.unpack(data)

    return buffer_size


def iter_blocks(f, thinning=False):
    is_fortran_file = True
    if not thinning:
        block_size = BLOCK_SIZE_BYTES
        buffer_size = DEFAULT_BUFFER_SIZE
    else:
        block_size = BLOCK_SIZE_BYTES_THIN
        buffer_size = DEFAULT_BUFFER_SIZE_THIN

    data = f.read(4)
    first = True
    if data == b"RUNH":
        is_fortran_file = False
    elif data in (MagicBytes.EVENTIO_LE.value, MagicBytes.EVENTIO_BE.value):
        msg = "This file is in eventio format, try opening with eventio.IACTFile"
        raise ValueError(msg)

    while True:
        # for the fortran-chunked output, we need to read the record size
        if is_fortran_file:
            if first is True:
                data = data + f.read(RECORD_MARKER.size - len(data))
            else:
                data = f.read(RECORD_MARKER.size)

            if len(data) == 0:
                return

            if len(data) < RECORD_MARKER.size:
                raise IOError("Read less bytes than expected, file seems to be truncated")

            buffer_size, = RECORD_MARKER.unpack(data)
            data = b""

        if first is True:
            data = data + f.read(buffer_size - len(data))
            first = False
        else:
            data = f.read(buffer_size)

        if is_fortran_file:
            if len(data) < buffer_size:
                raise IOError("Read less bytes than expected, file seems to be truncated")

        else:
            if len(data) == 0:
                return

        n_blocks, rest = divmod(len(data), block_size)
        if rest != 0:
            raise IOError("Read less bytes than expected, file seems to be truncated")

        for block in range(n_blocks):
            start = block * block_size
            stop = start + block_size
            block = data[start:stop]
            yield block

        # read trailing record marker
        if is_fortran_file:
            f.read(RECORD_MARKER.size)


def read_block(f, thinning=False, buffer_size=None):
    '''
    Reads a block of CORSIKA output, e.g. 273 4-byte floats.

    Under some conditions, CORSIKA writes output as FORTRAN
    raw file format. This means, there is a 4-byte unsigned
    integer (the buffer size)
    before and after each block, which has to be skipped.
    '''
    if thinning == False:
        block_size = BLOCK_SIZE_BYTES
    else:
        block_size = BLOCK_SIZE_BYTES_THIN
    if buffer_size is not None:
        pos = f.tell()
        if pos == 0:
            f.seek(4)

        if (pos + 4) % (buffer_size + 8) == 0:
            f.read(8)

    block = f.read(block_size)
    if len(block) < block_size:
        raise IOError("Read less bytes than expected, file seems to be truncated")

    return block



from functools import cache
import numpy as np

from .dtypes import Field, build_dtype, filter_fields_by_version


run_header_fields = [
    Field(1, "run_header", dtype="S4", min_version=np.float32(6.5)),
    Field(2, "run_number", min_version=np.float32(6.5)),
    Field(3, "date", min_version=np.float32(6.5)),
    Field(4, "version", min_version=np.float32(6.5)),
    Field(5, "n_observation_levels", min_version=np.float32(6.5)),
    Field(5 + 1, "observation_height", unit="cm", shape=10, min_version=np.float32(6.5)),
    Field(16, "energy_spectrum_slope", min_version=np.float32(6.5)),
    Field(17, "energy_min", unit="GeV", min_version=np.float32(6.5)),
    Field(18, "energy_max", unit="GeV", min_version=np.float32(6.5)),
    Field(19, "egs4_flag", min_version=np.float32(6.5)),
    Field(20, "nkg_flag", min_version=np.float32(6.5)),
    Field(21, "energy_cutoff_hadrons", unit="GeV", min_version=np.float32(6.5)),
    Field(22, "energy_cutoff_muons", unit="GeV", min_version=np.float32(6.5)),
    Field(23, "energy_cutoff_electrons", unit="GeV", min_version=np.float32(6.5)),
    Field(24, "energy_cutoff_photons", unit="GeV", min_version=np.float32(6.5)),
    Field(25, "physical_constants_and_interaction_flags", shape=50, min_version=np.float32(6.5)),
    Field(75, "inclined_observation_plane_x", unit="cm", min_version=np.float32(7.3)),
    Field(76, "inclined_observation_plane_y", unit="cm", min_version=np.float32(7.3)),
    Field(77, "inclined_observation_plane_z", unit="cm", min_version=np.float32(7.3)),
    Field(78, "inclined_observation_plane_theta", unit="deg", min_version=np.float32(7.3)),
    Field(79, "inclined_observation_plane_phi", unit="deg", min_version=np.float32(7.3)),
    Field(92, "rotation_angle_to_true_north", unit="deg", min_version=np.float32(7.75)),
    Field(93, "n_showers", min_version=np.float32(7.4)),  # typo in the 7.400 manual
    Field(94 + 1, "cka", shape=40, min_version=np.float32(6.5)),
    Field(134 + 1, "ceta", shape=5, min_version=np.float32(6.5)),
    Field(139 + 1, "cstrba", shape=11, min_version=np.float32(6.5)),
    Field(248, "x_scatter", unit="cm", min_version=np.float32(7.3)),
    Field(249, "y_scatter", unit="cm", min_version=np.float32(7.3)),
    Field(254 + 1, "aatm", shape=5, min_version=np.float32(6.5)),
    Field(259 + 1, "batm", shape=5, min_version=np.float32(6.5)),
    Field(264 + 1, "catm", shape=5, min_version=np.float32(6.5)),
    Field(270, "nflain", min_version=np.float32(6.5)),
    Field(271, "nfdif", min_version=np.float32(6.5)),
    Field(272, "nflpi0_100nflpif", min_version=np.float32(6.5)),
    Field(273, "nflche_100nfragm", min_version=np.float32(6.5)),
]


def get_run_header_fields(version):
    return filter_fields_by_version(run_header_fields, version)


@cache
def get_run_header_dtype(version):
    return build_dtype(get_run_header_fields(version))


@cache
def get_run_header_thin_dtype(version):
    return build_dtype(get_run_header_fields(version), itemsize=4 * 312)
